package xyz.twovb.sgm.games.impl.capturethebrick;

/*
 * Created by 2vb - 5/7/2024
 */

import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import net.kyori.adventure.text.minimessage.MiniMessage;
import net.kyori.adventure.text.minimessage.tag.resolver.Placeholder;
import net.kyori.adventure.text.minimessage.tag.resolver.TagResolver;
import net.kyori.adventure.text.serializer.legacy.LegacyComponentSerializer;
import net.kyori.adventure.text.serializer.plain.PlainTextComponentSerializer;
import net.megavex.scoreboardlibrary.api.sidebar.Sidebar;
import net.megavex.scoreboardlibrary.api.sidebar.component.ComponentSidebarLayout;
import net.megavex.scoreboardlibrary.api.sidebar.component.SidebarComponent;
import net.megavex.scoreboardlibrary.api.sidebar.component.animation.CollectionSidebarAnimation;
import net.megavex.scoreboardlibrary.api.sidebar.component.animation.SidebarAnimation;
import net.megavex.scoreboardlibrary.api.team.ScoreboardTeam;
import net.megavex.scoreboardlibrary.api.team.TeamDisplay;
import net.megavex.scoreboardlibrary.api.team.TeamManager;
import net.megavex.scoreboardlibrary.api.team.enums.CollisionRule;
import org.bukkit.*;
import org.bukkit.block.Block;
import org.bukkit.command.CommandSender;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.entity.ArmorStand;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.HandlerList;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityDismountEvent;
import org.bukkit.event.entity.PlayerDeathEvent;
import org.bukkit.event.player.*;
import org.bukkit.inventory.EntityEquipment;
import org.bukkit.inventory.ItemStack;
import org.bukkit.scheduler.BukkitRunnable;
import org.bukkit.scheduler.BukkitTask;
import org.bukkit.scoreboard.Scoreboard;
import org.jetbrains.annotations.NotNull;
import xyz.twovb.sgm.SGM;
import xyz.twovb.sgm.games.GameManager;
import xyz.twovb.sgm.games.Minigame;
import xyz.twovb.sgm.games.impl.capturethebrick.events.*;
import xyz.twovb.sgm.levels.LevelManager;
import xyz.twovb.toolbox.api.CustomPlayer;
import xyz.twovb.toolbox.utils.ChatUtils;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.*;

public class CTB implements Minigame {

    final static String name = "CaptureTheBrick";
    final static FileConfiguration messages = SGM.getInstance().getConfigManager().loadConfig("games/" + name.toLowerCase() + "/messages.yml");
    private final UUID uuid;
    private final ArrayList<Location> redBrickSpawns = new ArrayList<Location>();
    private final ArrayList<Location> blueBrickSpawns = new ArrayList<Location>();
    private Location redSpawnLocation;
    private Location blueSpawnLocation;
//    private final ArrayList<Location> redSpawnArea = new ArrayList<Location>();
//    private final ArrayList<Location> blueSpawnArea = new ArrayList<Location>();
    private final ArrayList<Player> players = new ArrayList<Player>();
    private final HashMap<Player, Location> playerLocationHashMap = new HashMap<>();
    // tagger, tagged
    private final HashMap<Player, Player> taggedPlayerHashMap = new HashMap<Player, Player>();
    private final HashMap<Player, CtbData.PlayerState> playerStateMap = new HashMap<>();
    ScoreboardTeam redTeam;
    ScoreboardTeam blueTeam;
    private Location redJail;
    private Location blueJail;
    private World gameWorld;
    private GameState state;
    private int bricks;
    private int territoryLevel;
    private Location spawnLoc;
    private TeamManager teamManager;
    private Sidebar sb;
    private SidebarAnimation<Component> headerAnimation;
    private ComponentSidebarLayout layout;

    private int tagTime;
    private int jailTime;

    private int taskId;

    public CTB() {
        this.uuid = UUID.randomUUID();
        this.state = GameState.PRESTART;
//        this.messages = SGM.getInstance().getConfigManager().loadConfig("games/" + name.toLowerCase() + "/messages.yml");
    }

    public static FileConfiguration getMessages() {
        return messages;
    }

    public void init(CommandSender owner, String world) throws IOException {
        // Perform initialization logic here
        gameWorld = createGameWorld(world, uuid);
        gameWorld.setAutoSave(false);
        if (applyOptions(world)) {
            this.teamManager = SGM.getInstance().getScoreboardLibrary().createTeamManager();
            this.redTeam = teamManager.createIfAbsent("red_team");
            this.blueTeam = teamManager.createIfAbsent("blue_team");
            Bukkit.getPluginManager().registerEvents(this, SGM.getInstance());

            setupTeamDisplay(redTeam, "Red Team", NamedTextColor.RED);
            setupTeamDisplay(blueTeam, "Blue Team", NamedTextColor.BLUE);

            this.sb = SGM.getInstance().getScoreboardLibrary().createSidebar();
            this.headerAnimation = createGradientAnimation(ChatUtils.translate("<bold>Capture The Brick"));
            var title = SidebarComponent.animatedLine(headerAnimation);

            SimpleDateFormat dtf = new SimpleDateFormat("dd/MM/yy | HH:mm:ss");

            SidebarComponent gameInfo = SidebarComponent.builder()
                    // @formatter:off
                    .addDynamicLine(() -> {
                        var time = dtf.format(new Date());
                        return Component.text(time, NamedTextColor.GRAY);
                    })
                    .addBlankLine()
                    .addDynamicLine(() -> ChatUtils.translate("<red>Red: <bold>" + getBrickCount(redTeam)))
                    .addDynamicLine(() -> ChatUtils.translate("<blue>Blue: <bold>" + getBrickCount(blueTeam)))
                    .addBlankLine()
                    .build();
            // @formatter:on

            this.layout = new ComponentSidebarLayout(title, gameInfo);
            layout.apply(sb);

            // Ready
            state = GameState.READY;
            for (Player player : Bukkit.getOnlinePlayers()) {
                CustomPlayer customPlayer = new CustomPlayer(player);
                customPlayer.sendMessage(messages.getString("system.game.ready").replace("%game%", name));
            }
        } else {
            owner.sendMessage(ChatUtils.translate(messages.getString("system.game.failed").replace("%cause%", "to load options.")));
            this.stop();
        }
    }

    private String getBrickCount(ScoreboardTeam team) {
        String icon = "\uD83D\uDDC3";
        StringBuilder builder = new StringBuilder();
        for (int i = 1; i <= this.bricks; i++) {
            builder.append(icon + " ");
        }
        return builder.toString();
    }

    @Override
    public void onTick() {
        if (players.isEmpty()) {
            this.stop();
        }
        try {
            headerAnimation.nextFrame();
            layout.apply(sb);
        } catch (IllegalStateException ignored) {
        }

        for (Player player : gameWorld.getPlayers()) {
            if (!players.contains(player)) {
//                for (Player gamePlayer : players) {
//                    gamePlayer.hidePlayer(SGM.getInstance(), player);
//                }
                sendMessageToAllPlayers(player.getName() + " IS A IMPOSTER");
            }
        }
        for (Player player : players) {
            CtbData.Territory territory = getTerritory(player);
            CtbData.PlayerState state = playerStateMap.get(player);
            Player tagged = taggedPlayerHashMap.get(player);
            switch (state) {
                case Tagging:
                    if (tagged != null) {
                        player.sendActionBar(ChatUtils.translate(state.getFancyText().replace("%player%", tagged.getName())));
                    } else {
                        player.sendActionBar(ChatUtils.translate(territory.getFancyText()));
                    }
                    break;
                case Jailed:
                    player.sendActionBar(ChatUtils.translate(state.getFancyText().replace("%territory%", territory.getFancyText())));
                    break;
                case Tagged, Invulnerable, Stealing:
                    player.sendActionBar(ChatUtils.translate(state.getFancyText()));
                    break;
                default:
                    player.sendActionBar(ChatUtils.translate(territory.getFancyText()));
                    break;
            }
        }
        SGM.getInstance().getCLogger().log(playerStateMap);
    }

    @Override
    public void start(CommandSender sender) {
        if (checkStart(sender)) {
            for (Player player : players) {
                player.getInventory().clear();
                player.teleport(this.spawnLoc);
                sb.addPlayer(player);
                player.setGameMode(GameMode.ADVENTURE);
                playerStateMap.put(player, CtbData.PlayerState.Free);
                for (ScoreboardTeam teams : teamManager.teams()) {

                }
            }
            for (int i = 0; i <= this.bricks - 1; i++) {
                placeBrick(redBrickSpawns.get(i), redTeam);
                placeBrick(blueBrickSpawns.get(i), blueTeam);
            }
            BukkitTask task = Bukkit.getScheduler().runTaskTimer(SGM.getInstance(), this::onTick, 0L, 1L);
            this.taskId = task.getTaskId();

            state = GameState.STARTED;
            sendMessageToAllPlayers("The game has started!");
        }
    }

    boolean isJailed(Player player) {
        return playerStateMap.get(player) == CtbData.PlayerState.Jailed;
    }

    CtbData.Territory getTerritory(Player player) {
        Location playerLoc = player.getLocation();
        Location blockCheck = new Location(gameWorld, playerLoc.getBlockX(), territoryLevel, playerLoc.getBlockZ());
        Block territoryCheck = blockCheck.getBlock();

        switch (territoryCheck.getType()) {
            case RED_CONCRETE:
                return CtbData.Territory.Red;
            case BLUE_CONCRETE:
                return CtbData.Territory.Blue;
            case WHITE_CONCRETE:
                return CtbData.Territory.Neutral;
            default:
                return CtbData.Territory.Unknown;
        }
    }

    // Check if we should be starting rn
    private boolean checkStart(CommandSender sender) {
        if (state != GameState.READY) return false;
        for (Player player : players) {
            if (!teamManager.players().contains(player)) {
                sender.sendMessage("All players need to be on a team!");
                return false;
            }
        }
        return true;
    }

    @Override
    public void stop() {
        // TODO: make it wait before unloading world
        // Stop game logic
        state = GameState.ENDING;
        if (!players.isEmpty()) {
            players.clear();
        }
        GameManager.getActiveGames().remove(uuid);
        HandlerList.unregisterAll(this);
        for (Player player : gameWorld.getPlayers()) {
            player.teleport(Bukkit.getServer().getWorlds().get(0).getSpawnLocation());
        }
        Bukkit.unloadWorld(gameWorld, false);
        if (teamManager != null) {
            teamManager.close();
        }
        if (sb != null) {
            sb.close();
        }
        Bukkit.getScheduler().cancelTask(this.taskId);
    }

    @Override
    public void addPlayer(Player player) {
        player.teleport(this.spawnLoc);
        players.add(player);
        sendMessageToAllPlayers(messages.getString("system.player.join").replace("%player%", player.getName()));
    }

    @Override
    public void removePlayer(Player player) {
        sendMessageToAllPlayers(messages.getString("system.player.left").replace("%player%", player.getName()));
        sb.removePlayer(player);
        removePlayerFromTeams(player);
        players.remove(player);
    }

    @EventHandler
    void onPlayerQuit(PlayerQuitEvent event) {
        Player player = event.getPlayer();
        if (players.contains(player)) {
            removePlayer(event.getPlayer());
        }
    }

    @EventHandler
    void onWorldChange(PlayerChangedWorldEvent event) {
        Player player = event.getPlayer();
        if (players.contains(player) && event.getFrom().equals(gameWorld)) {
            removePlayer(event.getPlayer());
        }
    }

    @EventHandler
    void onPlayerDeath(PlayerDeathEvent event) {
        Player player = event.getPlayer();
        if (players.contains(player) && state == GameState.STARTED) {
            sendMessageToAllPlayers(player.getName() + " was eliminated!");
            event.setDeathMessage("");
            removePlayer(player);
        }
    }

    @EventHandler
    void onPlayerSneak(PlayerToggleSneakEvent event) {
        Player tagger = event.getPlayer();
        if (!players.contains(tagger)) return;
        if (this.state != GameState.STARTED) return;
        if (playerStateMap.get(tagger) != CtbData.PlayerState.Tagging) return;
        Player tagged = taggedPlayerHashMap.get(tagger);
        if (tagged != null) {
            releasePlayer(tagged, tagger);
            for (Entity entity : tagger.getPassengers()) {
                if (entity instanceof ArmorStand) {
                    entity.remove();
                }
            }
        }
    }

    private void releasePlayer(Player tagged, Player tagger) {
        playerStateMap.replace(tagger, CtbData.PlayerState.Tagging, CtbData.PlayerState.Invulnerable);
        playerStateMap.replace(tagged, CtbData.PlayerState.Tagged, CtbData.PlayerState.Invulnerable);
        CustomPlayer cTagged = new CustomPlayer(tagged);
        CustomPlayer cTagger = new CustomPlayer(tagger);
        cTagged.sendMessage(messages.getString("system.player.released.info").replace("%player%", tagger.getName()));
        cTagger.sendMessage(messages.getString("system.player.released.notify").replace("%player%", tagged.getName()));
        taggedPlayerHashMap.remove(tagger);
    }

    @EventHandler
    void onPlayerMoveWithBrick(PlayerMoveEvent event) {
        if (!(event.getFrom().getX() != event.getTo().getX() || event.getFrom().getZ() != event.getTo().getZ())) return;
        Player player = event.getPlayer();
        if (!players.contains(player)) return;
        if (this.state != GameState.STARTED) return;
        if (playerStateMap.get(player) != CtbData.PlayerState.Stealing) return;
        CtbData.Territory territory = getTerritory(player);
//        CustomPlayer cPlayer = new CustomPlayer(player);
        if (territory.equals(CtbData.Territory.Neutral)) {
            sendMessageToAllPlayers(messages.getString("system.player.brick.steal.announce").replace("%player%", player.getName()));
            playerStateMap.replace(player, CtbData.PlayerState.Stealing, CtbData.PlayerState.Free);
        }
    }

    @EventHandler
    void onPlayerStealBrick(PlayerStealBrickEvent event) {
        Player player = event.getThief();
        if (!players.contains(player)) return;
        if (this.state != GameState.STARTED) return;
        if (playerStateMap.get(player) != CtbData.PlayerState.Stealing) return;
        player.sendMessage(event.getMessage());
    }

    @EventHandler
    void onPlayerWalkWhileInvuln(PlayerMoveEvent event) {
        if (!(event.getFrom().getX() != event.getTo().getX() || event.getFrom().getZ() != event.getTo().getZ())) return;
        Player player = event.getPlayer();
        if (!players.contains(player)) return;
        if (this.state != GameState.STARTED) return;
        if (playerStateMap.get(player) != CtbData.PlayerState.Invulnerable) return;
        CtbData.Territory territory = getTerritory(player);
        CustomPlayer cPlayer = new CustomPlayer(player);
        if (territory.equals(CtbData.Territory.Neutral)) {
            cPlayer.sendMessage(messages.getString("system.player.released.full"));
            playerStateMap.replace(player, CtbData.PlayerState.Invulnerable, CtbData.PlayerState.Free);
        }
    }

    @EventHandler
    void onPlayerWalkWhileTag(PlayerMoveEvent event) {
        if (!(event.getFrom().getX() != event.getTo().getX() || event.getFrom().getZ() != event.getTo().getZ())) return;
        Player tagger = event.getPlayer();
        if (!players.contains(tagger)) return;
        if (this.state != GameState.STARTED) return;
        if (playerStateMap.get(tagger) != CtbData.PlayerState.Tagging) return;
        ScoreboardTeam team = getTeam(tagger);
        Location jail = gameWorld.getSpawnLocation();
        if (team.equals(redTeam)) {
            jail = redJail;
        } else if (team.equals(blueTeam)) {
            jail = blueJail;
        }

        CtbData.Territory territory = getTerritory(tagger);

        Player tagged = taggedPlayerHashMap.get(tagger);
        if (tagged != null) {
            if (territory == CtbData.Territory.Neutral) {
                releasePlayer(tagged, tagger);
                for (Entity entity : tagger.getPassengers()) {
                    if (entity instanceof ArmorStand) {
                        entity.remove();
                    }
                }
            }
            double distanceToJail = jail.distanceSquared(tagger.getLocation());
            if (distanceToJail <= 12) {
                addPlayerToJail(tagged);
                taggedPlayerHashMap.remove(tagger);
                playerStateMap.replace(tagger, CtbData.PlayerState.Tagging, CtbData.PlayerState.Free);
                for (Entity entity : tagger.getPassengers()) {
                    if (entity instanceof ArmorStand) {
                        entity.remove();
                    }
                }
            }
        }
    }

    @EventHandler
    void onPlayerMoveInJail(PlayerMoveEvent event) {
        if (!(event.getFrom().getX() != event.getTo().getX() || event.getFrom().getZ() != event.getTo().getZ())) return;
        Player player = event.getPlayer();
        if (!players.contains(player)) return;
        if (this.state != GameState.STARTED) return;
        if (playerStateMap.get(player) != CtbData.PlayerState.Jailed) return;
        CtbData.Territory territory = getTerritory(player);
        Location jail = gameWorld.getSpawnLocation();
        if (territory.equals(CtbData.Territory.Blue)) {
            jail = this.blueJail;
        } else if (territory.equals(CtbData.Territory.Red)) {
            jail = this.redJail;
        }
        double distanceFromJail = jail.distanceSquared(player.getLocation());
        if (distanceFromJail >= 9) {
            Location lastValidPos = playerLocationHashMap.get(player);
            player.teleport(Objects.requireNonNullElse(lastValidPos, jail));
        } else {
            playerLocationHashMap.put(player, player.getLocation());
        }
    }

    @EventHandler
    void onPlayerTag(PlayerTaggedEvent event) {
        Player tagged = event.getTagged();
        Player tagger = event.getTagger();
        if (this.state != GameState.STARTED) return;
        if (!players.contains(tagged) || !players.contains(tagger)) return;
        if (playerStateMap.get(tagged) == CtbData.PlayerState.Stealing) {
            ScoreboardTeam team = getTeam(tagged);
            if (team == redTeam) {
                placeBrick(tagged.getLocation(), blueTeam);
            } else if (team == blueTeam) {
                placeBrick(tagged.getLocation(), redTeam);
            }
        }
        taggedPlayerHashMap.putIfAbsent(tagger, tagged);
        CustomPlayer cTagged = new CustomPlayer(tagged);
        CustomPlayer cTagger = new CustomPlayer(tagger);
        cTagged.sendMessage(messages.getString("system.player.tag.notify").replace("%player%", tagger.getName()));
        cTagger.sendMessage(messages.getString("system.player.tag.info").replace("%player%", tagged.getName()));
        Entity seatEntity = gameWorld.spawnEntity(tagger.getLocation(), EntityType.ARMOR_STAND);
        ArmorStand armorStandEntity = (ArmorStand) seatEntity;
        armorStandEntity.setInvisible(true);
        armorStandEntity.setSmall(true);
        tagger.addPassenger(armorStandEntity);
        armorStandEntity.addPassenger(tagged);
        new BukkitRunnable() {
            private int seconds = 0;
            @Override
            public void run() {
                if (seconds > tagTime) {
                    tagger.sendMessage("out of time");
                } else {
                    tagger.sendMessage(String.valueOf(tagTime));
                    seconds++;
                }
            }
        }.runTaskTimer(SGM.getInstance(), 0, 20L);
    }

    @EventHandler
    void onPlayerFreedom(PlayerFreedFromJailEvent event) {
        Player tagged = event.getTagged();
        Player tagger = event.getTagger();
        if (this.state != GameState.STARTED) return;
        if (!players.contains(tagged) || !players.contains(tagger)) return;
        CustomPlayer cTagged = new CustomPlayer(tagged);
        CustomPlayer cTagger = new CustomPlayer(tagger);
        sendMessageToAllPlayers(messages.getString("system.player.jail.free.notify").replace("%rescuer%", tagger.getName()).replace("%player%", tagged.getName()));
        cTagger.sendMessage(messages.getString("system.player.jail.free.info").replace("%player%", tagged.getName()));
        cTagged.sendMessage(messages.getString("system.player.jail.free.info").replace("%player%", tagger.getName()));
    }

    @EventHandler
    void onDismountAttempt(EntityDismountEvent event) {
        if (!(event.getEntity() instanceof Player player)) return;
        if (this.state != GameState.STARTED) return;
        if (!players.contains(player)) return;
        CtbData.PlayerState playerState = playerStateMap.get(player);
        if (playerState != CtbData.PlayerState.Tagged) return;
        event.setCancelled(true);
    }

    @EventHandler
    void onPlayerJail(PlayerJailedEvent event) {
        Player player = event.getJailed();
        if (this.state != GameState.STARTED) return;
        if (!players.contains(player)) return;
        if (playerStateMap.get(player) != CtbData.PlayerState.Jailed) return;
        CustomPlayer cPlayer = new CustomPlayer(player);
        cPlayer.sendMessage(messages.getString("system.player.jail.info"));
        for (String teamMembers : getTeam(cPlayer.player()).defaultDisplay().entries()) {
            Player teamPlayer = Bukkit.getServer().getPlayerExact(teamMembers);
            CustomPlayer teamMember = new CustomPlayer(teamPlayer);
            if (teamMember.player() == cPlayer.player()) continue;
            teamMember.sendMessage(LegacyComponentSerializer.legacyAmpersand().serialize(event.getMessage()));
        }
    }



    @EventHandler
    void handleEntityDamage(EntityDamageByEntityEvent event) {
        if (this.state != GameState.STARTED) return;
        Entity eventEntity = event.getEntity();
        Entity eventDamager = event.getDamager();

        if (!(eventEntity instanceof Player interacted) || !(eventDamager instanceof Player interacter)) return;
        if (!players.contains(interacted) || !players.contains(interacter)) return;
//        if (getTerritory(interacted) != getTerritory(interacter)) return;

        PlayerGameInteractionEvent interactionEvent = new PlayerGameInteractionEvent(interacted, interacter);
        Bukkit.getServer().getPluginManager().callEvent(interactionEvent);

        event.setCancelled(true);
//        if (interactionEvent.isCancelled()) {
//            event.setCancelled(true);
//        }
    }

    @EventHandler
    void onPlayerFree(PlayerGameInteractionEvent event) {
        Player interacted = event.getInteracted();
        Player interacter = event.getInteracter();

        if (this.state != GameState.STARTED) return;
        if (!players.contains(interacted) || !players.contains(interacter)) return;
        if (getTerritory(interacted) != getTerritory(interacter)) return;
        ScoreboardTeam interactedTeam = getTeam(interacted);
        ScoreboardTeam interacterTeam = getTeam(interacter);
        if (interactedTeam != interacterTeam) return;
        if (playerStateMap.get(interacted) != CtbData.PlayerState.Jailed) return;
        if (playerStateMap.get(interacter) != CtbData.PlayerState.Free) return;

        playerStateMap.replace(interacted, CtbData.PlayerState.Jailed, CtbData.PlayerState.Invulnerable);
        playerStateMap.replace(interacter, CtbData.PlayerState.Free, CtbData.PlayerState.Invulnerable);
        PlayerFreedFromJailEvent freedFromJailEvent = new PlayerFreedFromJailEvent(interacted, interacter);
        Bukkit.getServer().getPluginManager().callEvent(freedFromJailEvent);
        event.setCancelled(true);
    }

    @EventHandler
    void onPlayerTag(PlayerGameInteractionEvent event) {
        Player interacted = event.getInteracted();
        Player interacter = event.getInteracter();

        if (this.state != GameState.STARTED) return;
        if (!players.contains(interacted) || !players.contains(interacter)) return;
        if (getTerritory(interacted) != getTerritory(interacter)) return;
        ScoreboardTeam interactedTeam = getTeam(interacted);
        ScoreboardTeam interacterTeam = getTeam(interacter);
        if (interactedTeam == interacterTeam) return;
        event.setCancelled(true);

        CtbData.Territory territory = getTerritory(interacted);
        if (territory == CtbData.Territory.Blue && interacterTeam != blueTeam) return;
        if (territory == CtbData.Territory.Red && interacterTeam != redTeam) return;
        if (playerStateMap.get(interacter) != CtbData.PlayerState.Free) return;
        if (playerStateMap.get(interacted) != CtbData.PlayerState.Free && playerStateMap.get(interacted) != CtbData.PlayerState.Stealing) return;

        playerStateMap.replace(interacted, CtbData.PlayerState.Tagged);
        playerStateMap.replace(interacter, CtbData.PlayerState.Free, CtbData.PlayerState.Tagging);
        PlayerTaggedEvent taggedEvent = new PlayerTaggedEvent(interacted, interacter);
        Bukkit.getServer().getPluginManager().callEvent(taggedEvent);
    }

    @EventHandler
    void onBrickSteal(PlayerInteractAtEntityEvent event) {
        // Check if a player is allowed to tag, and if the tagged player is allowed to be tagged
        // If both are true, call the custom tag event and add players to tagged hashmap
        if (this.state != GameState.STARTED) return;
        Entity rightClicked = event.getRightClicked();
        Player brickTaker = event.getPlayer();
        if (!(rightClicked instanceof ArmorStand)) return;
        if (!players.contains(brickTaker)) return;
        EntityEquipment equipment = ((ArmorStand) rightClicked).getEquipment();
        Material brickMaterial = null;
        if (equipment.getHelmet().getType().equals(Material.RED_WOOL)) {
            brickMaterial = Material.RED_WOOL;
        } else if (equipment.getHelmet().getType().equals(Material.BLUE_WOOL)) {
            brickMaterial = Material.BLUE_WOOL;
        }
        ScoreboardTeam taggerTeam = getTeam(brickTaker);
        event.setCancelled(true);
        CtbData.Territory territory = getTerritory(brickTaker);
        if (territory == CtbData.Territory.Blue && taggerTeam == blueTeam) return;
        if (territory == CtbData.Territory.Red && taggerTeam == redTeam) return;
        if (playerStateMap.get(brickTaker) != CtbData.PlayerState.Free) return;
        if (brickMaterial != null) return;
        playerStateMap.replace(brickTaker, CtbData.PlayerState.Free, CtbData.PlayerState.Stealing);
        EntityEquipment playerEquipment = event.getPlayer().getEquipment();
        playerEquipment.setHelmet(new ItemStack(brickMaterial));
        PlayerStealBrickEvent playerStealBrickEvent = new PlayerStealBrickEvent(brickTaker);
        playerStealBrickEvent.callEvent();
        rightClicked.remove();
    }

    public void addPlayerToJail(Player player) {
        CtbData.PlayerState playerState = playerStateMap.get(player);
        playerStateMap.replace(player, playerState, CtbData.PlayerState.Jailed);
        PlayerJailedEvent jailedEvent = new PlayerJailedEvent(player);
        jailedEvent.callEvent();
        ScoreboardTeam team = getTeam(player);
        if (team == redTeam) {
            player.teleport(blueJail);
        } else if (team == blueTeam) {
            player.teleport(redJail);
        }
    }

    public void freePlayerFromJail(Player player) {
        playerStateMap.replace(player, CtbData.PlayerState.Jailed, CtbData.PlayerState.Invulnerable);
    }

    @Override
    public List<Player> getPlayers() {
        return players;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public UUID getGameId() {
        return uuid;
    }

    @Override
    public GameState getState() {
        return state;
    }

    @Override
    public World getWorld() {
        return gameWorld;
    }

    public void addPlayerToTeam(Player player, CtbData.Teams team) {
        if (teamManager.players().contains(player)) {
            getTeam(player).defaultDisplay().removeEntry(player.getName());
        }
        if (team.equals(CtbData.Teams.Blue)) {
            blueTeam.defaultDisplay().addEntry(player.getName());
        } else if (team.equals(CtbData.Teams.Red)) {
            redTeam.defaultDisplay().addEntry(player.getName());
        }
        this.teamManager.addPlayer(player);
    }

    public void removePlayerFromTeams(Player player) {
        if (getTeam(player) != null) {
            getTeam(player).defaultDisplay().removeEntry(player.getName());
            this.teamManager.removePlayer(player);
        }
    }

    public ScoreboardTeam getTeam(Player player) {
        for (ScoreboardTeam team : teamManager.teams()) {
            if (team.defaultDisplay().entries().contains(player.getName())) {
                return team;
            }
        }
        return null;
    }

    private void placeBrick(Location location, ScoreboardTeam team) {
        Location brickSpawnLoc = location;
        brickSpawnLoc = brickSpawnLoc.clone().add(0.5, 0, 0.5);
        Entity brickEntity = gameWorld.spawnEntity(brickSpawnLoc, EntityType.ARMOR_STAND);
        ArmorStand armorStand = (ArmorStand) brickEntity;
        EntityEquipment armorEquipment = armorStand.getEquipment();
        // TODO: make sure the children cannot uneqip the wool :sob:
        if (team.equals(redTeam)) {
            armorEquipment.setHelmet(new ItemStack(Material.RED_WOOL));
        } else if (team.equals(blueTeam)) {
            armorEquipment.setHelmet(new ItemStack(Material.BLUE_WOOL));
        }
    }

    private void setupTeamDisplay(ScoreboardTeam team, String displayName, NamedTextColor color) {
        TeamDisplay display = team.defaultDisplay();
        display.displayName(Component.text(displayName));
        display.collisionRule(CollisionRule.NEVER);
        display.playerColor(color);
    }

    private @NotNull SidebarAnimation<Component> createGradientAnimation(@NotNull Component text) {
        float step = 1f / 8f;
        TagResolver.Single textPlaceholder = Placeholder.component("text", text);
        List<Component> frames = new ArrayList<>((int) (2f / step));
        float phase = -1f;

        while (phase <= 1) {
            String gradientTag = String.format("<gradient:#9bf8ad:#00ff04:%.2f>", phase);
            Component frame = MiniMessage.miniMessage().deserialize(gradientTag + "<text>", textPlaceholder);
            frames.add(frame);
            phase += step;
        }
        return new CollectionSidebarAnimation<>(frames);
    }

    private @NotNull SidebarAnimation<Component> createScrollingAnimation(@NotNull Component text, int frameDelay) {
        String rawText = PlainTextComponentSerializer.plainText().serialize(text);
        int textLength = rawText.length();
        int frameWidth = Math.min(20, textLength); // Use a maximum frame width of 20 or text length, whichever is smaller.

        // Determine the number of frames needed to scroll the text
        int totalFrames = textLength > frameWidth ? textLength - frameWidth + 1 : 1;
        List<Component> frames = new ArrayList<>(totalFrames * frameDelay);

        for (int i = 0; i < totalFrames; i++) {
            int start = i;
            int end = Math.min(i + frameWidth, textLength);
            String frameText = rawText.substring(start, end);

            // Add padding if the frame is shorter than frameWidth
            if (frameText.length() < frameWidth) {
                frameText = frameText + " ".repeat(frameWidth - frameText.length());
            }

            // Add multiple identical frames to slow down the animation
            Component frameComponent = Component.text(frameText, text.style());
            for (int j = 0; j < frameDelay; j++) {
                frames.add(frameComponent);
            }
        }

        return new CollectionSidebarAnimation<>(frames);
    }

    private Location parseString(String string) {
        String[] coords = string.trim().split(",");
        if (coords.length == 3) {
            return new Location(gameWorld, Double.parseDouble(coords[0]), Double.parseDouble(coords[1]), Double.parseDouble(coords[2]));
        } else {
            return null;
        }
    }

    // Attempts to load the game's options as a YamlConfiguration
    // before processing each section in the same order as the file.
    private boolean applyOptions(String worldName) {
        File configFile = new File(LevelManager.mapPath + worldName + "/sgm.yml");
        SGM.getInstance().getCLogger().log(configFile.getPath());
        SGM.getInstance().getCLogger().log(configFile.getAbsolutePath());
        YamlConfiguration config = YamlConfiguration.loadConfiguration(configFile);

        // Set map information
        ConfigurationSection mapInfoSection = config.getConfigurationSection("mapInfo");
        if (mapInfoSection != null) {
            this.bricks = mapInfoSection.getInt("brickCount");
            this.territoryLevel = mapInfoSection.getInt("territory");
            this.spawnLoc = parseString(mapInfoSection.getString("spawn"));
            this.jailTime = mapInfoSection.getInt("jailTime");
            this.tagTime = mapInfoSection.getInt("tagTime");
        } else {
            SGM.getInstance().getCLogger().error("Invalid map info. Please fix before trying again.");
            return false;
        }

        // Process team information

        CtbData.Teams[] teamsList = CtbData.Teams.values();
        for (CtbData.Teams team : teamsList) {
            String teamName = team.name().toLowerCase();
            ConfigurationSection teamConfig = config.getConfigurationSection("teams." + teamName);
            if (teamConfig != null) {
                List<String> brickSpawns = teamConfig.getStringList("brickSpawns");

                if (team.equals(CtbData.Teams.Red)) {
                    this.redJail = parseString(teamConfig.getString("jail"));
                    this.redSpawnLocation = parseString(teamConfig.getString("playerSpawn"));
                } else if (team.equals(CtbData.Teams.Blue)) {
                    this.blueJail = parseString(teamConfig.getString("jail"));
                    this.blueSpawnLocation = parseString(teamConfig.getString("playerSpawn"));
                }

                // Get brick locations
                if (brickSpawns.size() >= this.bricks) {
                    for (String rawBrickLoc : brickSpawns) {
                        Location brickLoc = parseString(rawBrickLoc);
                        if (team.equals(CtbData.Teams.Red)) {
                            this.redBrickSpawns.add(brickLoc);
                        } else if (team.equals(CtbData.Teams.Blue)) {
                            this.blueBrickSpawns.add(brickLoc);
                        }
                    }
                } else {
                    SGM.getInstance().getCLogger().error("Invalid brick spawns. Please fix before trying again.");
                    return false;
                }
            } else {
                SGM.getInstance().getCLogger().error("Invalid teams. Please fix before trying again.");
                return false;
            }
        }
        return true;
    }

}